@@ -42,6 +42,7 @@
 #ifdef HAVE_SYS_SOCKIO_H
 #include <sys/sockio.h>
 #endif
+#include <X11/Xlib.h>
 
 #include <vicious.h>
 
@@ -69,6 +70,22 @@ extern gboolean no_console;
 
 extern char *gdm_charset;
 
+#ifdef ENABLE_IPV6
+static gboolean have_ipv6 (void)
+{
+	int s;
+
+	s = socket (AF_INET6, SOCK_STREAM, 0);
+	if (s != -1) {
+		VE_IGNORE_EINTR (close (s));
+		return TRUE;
+	}
+
+	return FALSE;
+}
+#endif
+
+
 static void
 do_syslog (int type, const char *s)
 {
@@ -232,7 +249,7 @@ gdm_fdprintf (int fd, const gchar *format, ...)
 	written = 0;
 	while (written < len) {
 		int w;
-		IGNORE_EINTR (w = write (fd, &s[written], len - written));
+		VE_IGNORE_EINTR (w = write (fd, &s[written], len - written));
 		if (w < 0)
 			/* evil! */
 			break;
@@ -375,45 +392,65 @@ gdm_get_free_display (int start, uid_t server_uid)
 			continue;
 		}
 
-		sock = socket (AF_INET, SOCK_STREAM, 0);
+#ifdef ENABLE_IPV6
+		if (have_ipv6 ()) {
+			struct sockaddr_in6 serv6_addr= {0};
 
-		serv_addr.sin_port = htons (6000 + i);
+			sock = socket (AF_INET6, SOCK_STREAM,0);
 
-		errno = 0;
-		IGNORE_EINTR (connect (sock,
+			serv6_addr.sin6_family = AF_INET6;
+			serv6_addr.sin6_addr = in6addr_loopback;
+			serv6_addr.sin6_port = htons (6000 + i);
+			errno = 0;
+			VE_IGNORE_EINTR (connect (sock,
+                                      (struct sockaddr *)&serv6_addr,
+                                      sizeof (serv6_addr)));
+		}
+		else
+#endif
+		{
+			sock = socket (AF_INET, SOCK_STREAM, 0);
+
+			serv_addr.sin_port = htons (6000 + i);
+
+			errno = 0;
+			VE_IGNORE_EINTR (connect (sock,
 				       (struct sockaddr *)&serv_addr,
 				       sizeof (serv_addr)));
+		}
 		if (errno != 0 && errno != ECONNREFUSED) {
-			IGNORE_EINTR (close (sock));
+			VE_IGNORE_EINTR (close (sock));
 			continue;
 		}
-		IGNORE_EINTR (close (sock));
+		VE_IGNORE_EINTR (close (sock));
 
 		/* if lock file exists and the process exists */
 		g_snprintf (buf, sizeof (buf), "/tmp/.X%d-lock", i);
-		IGNORE_EINTR (r = stat (buf, &s));
+		VE_IGNORE_EINTR (r = stat (buf, &s));
 		if (r == 0 &&
 		    ! S_ISREG (s.st_mode)) {
 			/* Eeeek! not a regular file?  Perhaps someone
 			   is trying to play tricks on us */
 			continue;
 		}
-		fp = fopen (buf, "r");
+		VE_IGNORE_EINTR (fp = fopen (buf, "r"));
 		if (fp != NULL) {
 			char buf2[100];
-			if (fgets (buf2, sizeof (buf2), fp) != NULL) {
+			char *getsret;
+			VE_IGNORE_EINTR (getsret = fgets (buf2, sizeof (buf2), fp));
+			if (getsret != NULL) {
 				gulong pid;
 				if (sscanf (buf2, "%lu", &pid) == 1 &&
 				    kill (pid, 0) == 0) {
-					fclose (fp);
+					VE_IGNORE_EINTR (fclose (fp));
 					continue;
 				}
 
 			}
-			fclose (fp);
+			VE_IGNORE_EINTR (fclose (fp));
 
 			/* whack the file, it's a stale lock file */
-			IGNORE_EINTR (unlink (buf));
+			VE_IGNORE_EINTR (unlink (buf));
 		}
 
 		/* if starting as root, we'll be able to overwrite any
@@ -422,7 +459,7 @@ gdm_get_free_display (int start, uid_t server_uid)
 		if (server_uid > 0) {
 			g_snprintf (buf, sizeof (buf),
 				    "/tmp/.X11-unix/X%d", i);
-			IGNORE_EINTR (r = stat (buf, &s));
+			VE_IGNORE_EINTR (r = stat (buf, &s));
 			if (r == 0 &&
 			    s.st_uid != server_uid) {
 				continue;
@@ -430,7 +467,7 @@ gdm_get_free_display (int start, uid_t server_uid)
 
 			g_snprintf (buf, sizeof (buf),
 				    "/tmp/.X%d-lock", i);
-			IGNORE_EINTR (r = stat (buf, &s));
+			VE_IGNORE_EINTR (r = stat (buf, &s));
 			if (r == 0 &&
 			    s.st_uid != server_uid) {
 				continue;
@@ -571,7 +608,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 			return FALSE;
 		}
 
-		IGNORE_EINTR (close (tempfd));
+		VE_IGNORE_EINTR (close (tempfd));
 
 		argv[0] = EXPANDED_LIBEXECDIR "/gdmopen";
 		argv[1] = "-l";
@@ -596,19 +633,19 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 		g_free (argv[4]);
 
 		if (ret != NULL) {
-			fp = fopen (tempname, "r");
+			VE_IGNORE_EINTR (fp = fopen (tempname, "r"));
 			if (fp != NULL) {
 				if (fgets (buf, sizeof (buf), fp) != NULL &&
 				    (buf[0] == 'y' || buf[0] == 'Y'))
 					*ret = TRUE;
-				fclose (fp);
+				VE_IGNORE_EINTR (fclose (fp));
 			} else {
 				g_free (msg_quoted);
 				return FALSE;
 			}
 		}
 
-		IGNORE_EINTR (unlink (tempname));
+		VE_IGNORE_EINTR (unlink (tempname));
 
 		g_free (msg_quoted);
 		return TRUE;
@@ -650,7 +687,7 @@ gdm_exec_wait (char * const *argv, gboolean no_display,
 			ve_unsetenv ("XAUTHORITY");
 		}
 		
-		IGNORE_EINTR (execv (argv[0], argv));
+		VE_IGNORE_EINTR (execv (argv[0], argv));
 
 		_exit (-1);
 	}
@@ -818,8 +855,7 @@ gdm_ensure_sanity (void)
 	old_euid = geteuid ();
 	old_egid = getegid ();
 
-	NEVER_FAILS_seteuid (0);
-	NEVER_FAILS_setegid (0);
+	NEVER_FAILS_root_set_euid_egid (0, 0);
 
 	/* The /tmp/.ICE-unix check, note that we do
 	 * ignore errors, since it's not deadly to run
@@ -832,22 +868,21 @@ gdm_ensure_sanity (void)
 		   correct permissions */
 		struct stat s;
 		int r;
-		IGNORE_EINTR (r = lstat ("/tmp/.ICE-unix", &s));
+		VE_IGNORE_EINTR (r = lstat ("/tmp/.ICE-unix", &s));
 		if G_LIKELY (r == 0 && S_ISDIR (s.st_mode)) {
 			/* Make sure it is root and sticky */
-			IGNORE_EINTR (chown ("/tmp/.ICE-unix", 0, 0));
-			IGNORE_EINTR (chmod ("/tmp/.ICE-unix", 01777));
+			VE_IGNORE_EINTR (chown ("/tmp/.ICE-unix", 0, 0));
+			VE_IGNORE_EINTR (chmod ("/tmp/.ICE-unix", 01777));
 		} else {
 			/* There is a file/link/whatever called .ICE-unix? whack and try mkdir */
-			IGNORE_EINTR (unlink ("/tmp/.ICE-unix"));
+			VE_IGNORE_EINTR (unlink ("/tmp/.ICE-unix"));
 			mkdir ("/tmp/.ICE-unix", 01777);
 		}
 	}
 
 	umask (old_umask);
 
-	NEVER_FAILS_seteuid (old_euid);
-	NEVER_FAILS_setegid (old_egid);
+	NEVER_FAILS_root_set_euid_egid (old_euid, old_egid);
 }
 
 const GList *
@@ -855,18 +890,22 @@ gdm_peek_local_address_list (void)
 {
 	static GList *the_list = NULL;
 	static time_t last_time = 0;
-	struct in_addr *addr;
-#ifdef SIOCGIFCONF
 	struct sockaddr_in *sin;
+#ifdef ENABLE_IPV6
+	char hostbuf[BUFSIZ];
+	struct sockaddr_in6 *sin6;
+	struct addrinfo hints, *result, *res;
+#else /* ENABLE_IPV6 */
+	int sockfd;
+#ifdef SIOCGIFCONF
 	struct ifconf ifc;
 	struct ifreq *ifr;
 	char *buf;
 	int num;
-	int sockfd;
 #else /* SIOCGIFCONF */
 	char hostbuf[BUFSIZ];
 	struct hostent *he;
-	int i;
+#endif
 #endif
 
 	/* don't check more then every 5 seconds */
@@ -878,13 +917,51 @@ gdm_peek_local_address_list (void)
 	the_list = NULL;
 
 	last_time = time (NULL);
+#ifdef ENABLE_IPV6
+	hostbuf[BUFSIZ-1] = '\0';
+	if (gethostname (hostbuf, BUFSIZ-1) != 0) {
+		gdm_debug ("%s: Could not get server hostname", "gdm_peek_local_address_list");
+
+		sin6 = g_new0 (struct sockaddr_in6, 1);
+		sin6->sin6_family = AF_INET6;
+		sin6->sin6_addr = in6addr_loopback;
+		return g_list_append (the_list, sin6);
+	}
+
+	if (getaddrinfo (hostbuf, NULL, &hints, &result) != 0) {
+		gdm_debug ("%s: Could not get address from hostname!", "gdm_peek_local_address_list");
+
+		sin6 = g_new0 (struct sockaddr_in6, 1);
+		sin6->sin6_family = AF_INET6;
+		sin6->sin6_addr = in6addr_loopback;
+		return g_list_append (the_list, sin6);
+	}
+
+	for (res = result; res; res = res->ai_next) {
+		if (res->ai_family == AF_INET6) {
+			sin6 = g_new0 (struct sockaddr_in6, 1);
+			sin6->sin6_family = AF_INET6;
+			memcpy (sin6, res->ai_addr, res->ai_addrlen);
+			the_list = g_list_append (the_list, sin6);
+               }
+               else if (res->ai_family == AF_INET) {
+			sin = g_new0 (struct sockaddr_in, 1);
+			sin->sin_family = AF_INET;
+			memcpy (sin, res->ai_addr, res->ai_addrlen);
+			the_list = g_list_append (the_list, sin);
+               }
+	}
+
+	if (result) {
+		freeaddrinfo (result);
+		result = NULL;
+	}
+
+#else  /* ENABLE_IPV6 */
 
 #ifdef SIOCGIFCONF
-	if (gdm_xdmcpfd > 0)
-		sockfd = gdm_xdmcpfd;
-	else
-		/* Open a bogus socket */
-		sockfd = socket (AF_INET, SOCK_DGRAM, 0); /* UDP */
+/* Create an IPv4 socket to pick IPv4 addresses */
+	sockfd = socket (AF_INET, SOCK_DGRAM, 0); /* UDP */
 
 #ifdef SIOCGIFNUM
 	if (ioctl (sockfd, SIOCGIFNUM, &num) < 0) {
@@ -900,16 +977,18 @@ gdm_peek_local_address_list (void)
 		gdm_error (_("%s: Cannot get local addresses!"),
 			   "gdm_peek_local_address_list");
 		g_free (buf);
-		if (sockfd != gdm_xdmcpfd)
-			IGNORE_EINTR (close (sockfd));
-		addr = g_new0 (struct in_addr, 1);
-		addr->s_addr = INADDR_LOOPBACK;
-		return g_list_append (NULL, addr);
+		VE_IGNORE_EINTR (close (sockfd));
+		sin = g_new0 (struct sockaddr_in, 1);
+		sin->sin_family = AF_INET;
+		sin->sin_addr.s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, sin);
 	}
 
 	ifr = ifc.ifc_req;
 	num = ifc.ifc_len / sizeof(struct ifreq);
 	for (; num-- > 0; ifr++) {
+		struct sockaddr_in *addr;
+
 		if (ioctl (sockfd, SIOCGIFFLAGS, ifr) < 0 ||
 		    ! (ifr->ifr_flags & IFF_UP))
 			continue;
@@ -926,48 +1005,71 @@ gdm_peek_local_address_list (void)
 		    sin->sin_addr.s_addr == INADDR_ANY ||
 		    sin->sin_addr.s_addr == INADDR_LOOPBACK)
 			continue;
-		addr = g_new0 (struct in_addr, 1);
-		memcpy (addr, &(sin->sin_addr.s_addr),
-			sizeof (struct in_addr));
+		addr = g_new0 (struct sockaddr_in, 1);
+		memcpy (addr, sin, sizeof (struct sockaddr_in));
 		the_list = g_list_append (the_list, addr);
 	}
 
-	if (sockfd != gdm_xdmcpfd)
-		IGNORE_EINTR (close (sockfd));
-
+		VE_IGNORE_EINTR (close (sockfd));
 	g_free (buf);
 #else /* SIOCGIFCONF */
 	/* host based fallback, will likely only get 127.0.0.1 i think */
 
 	hostbuf[BUFSIZ-1] = '\0';
+	sin = g_new0 (struct sockaddr_in, 1);
+	sin->sin_family = AF_INET;
 	if G_UNLIKELY (gethostname (hostbuf, BUFSIZ-1) != 0) {
 		gdm_debug ("%s: Could not get server hostname: %s!",
 			   "gdm_peek_local_address_list",
 			   strerror (errno));
-		addr = g_new0 (struct in_addr, 1);
-		addr->s_addr = INADDR_LOOPBACK;
-		return g_list_append (NULL, addr);
+		sin->sin_addr.s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, sin);
 	}
 	he = gethostbyname (hostbuf);
 	if G_UNLIKELY (he == NULL) {
 		gdm_debug ("%s: Could not get address from hostname!",
 			   "gdm_peek_local_address_list");
-		addr = g_new0 (struct in_addr, 1);
-		addr->s_addr = INADDR_LOOPBACK;
-		return g_list_append (NULL, addr);
+		sin->sin_addr.s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, sin);
 	}
 	for (i = 0; he->h_addr_list[i] != NULL; i++) {
 		struct in_addr *laddr = (struct in_addr *)he->h_addr_list[i];
 
-		addr = g_new0 (struct in_addr, 1);
-		memcpy (addr, laddr, sizeof (struct in_addr));
-		the_list = g_list_append (the_list, addr);
+		memcpy (&sin->sin_addr, laddr, sizeof (struct in_addr));
+		the_list = g_list_append (the_list, sin);
 	}
 #endif
+#endif /* ENABLE_IPV6 */
 
 	return the_list;
 }
 
+#ifdef ENABLE_IPV6
+gboolean
+gdm_is_local_addr6 (struct in6_addr* ia)
+{
+
+	if (IN6_IS_ADDR_LOOPBACK (ia)) {
+		return TRUE;
+
+	} else {
+		const GList *list = gdm_peek_local_address_list ();
+
+		while (list != NULL) {
+			struct sockaddr *addr = list->data;
+
+			if ((addr->sa_family == AF_INET6) && (memcmp (ia, &(((struct sockaddr_in6 *)addr)->sin6_addr), sizeof (struct in6_addr)) == 0)) {
+				return TRUE;
+			}
+
+			list = list->next;
+		}
+
+		return FALSE;
+	}
+}
+#endif
+
 gboolean
 gdm_is_local_addr (struct in_addr *ia)
 {
@@ -980,9 +1082,10 @@ gdm_is_local_addr (struct in_addr *ia)
 		const GList *list = gdm_peek_local_address_list ();
 
 		while (list != NULL) {
-			struct in_addr *addr = list->data;
+			struct sockaddr *addr = list->data;
 
-			if (memcmp (&ia->s_addr, &addr->s_addr, 4) == 0) {
+			if ((addr->sa_family == AF_INET ) && 
+			    (memcmp (ia, &(((struct sockaddr_in *)addr)->sin_addr), 4) == 0)) {
 				return TRUE;
 			}
 
@@ -993,6 +1096,17 @@ gdm_is_local_addr (struct in_addr *ia)
 	}
 }
 
+#ifdef ENABLE_IPV6
+gboolean
+gdm_is_loopback_addr6 (struct in6_addr *ia)
+{
+	if (IN6_IS_ADDR_LOOPBACK(ia))
+		return TRUE;
+
+	return FALSE;
+}
+#endif
+
 gboolean
 gdm_is_loopback_addr (struct in_addr *ia)
 {
@@ -1034,12 +1148,12 @@ gdm_desetuid (void)
 	{
 		int setresuid(uid_t ruid, uid_t euid, uid_t suid);
 		int setresgid(gid_t rgid, gid_t egid, gid_t sgid);
-		setresuid (uid, uid, uid);
 		setresgid (gid, gid, gid);
+		setresuid (uid, uid, uid);
 	}
 #else
-	seteuid (getuid ());
 	setegid (getgid ());
+	seteuid (getuid ());
 #endif
 }
 
@@ -1112,7 +1226,7 @@ gdm_test_opt (const char *cmd, const char *help, const char *option)
 
 		got_it = TRUE;
 	}
-	fclose (fp);
+	VE_IGNORE_EINTR (fclose (fp));
 	return got_it;
 }
 
@@ -1122,7 +1236,7 @@ gdm_fdgetc (int fd)
 	char buf[1];
 	int bytes;
 
-	IGNORE_EINTR (bytes = read (fd, buf, 1));
+	VE_IGNORE_EINTR (bytes = read (fd, buf, 1));
 	if (bytes != 1)
 		return EOF;
 	else
@@ -1179,7 +1293,7 @@ gdm_close_all_descriptors (int from, int except, int except2)
 		closedir (dir);
 		for (li = openfds; li != NULL; li = li->next) {
 			int fd = GPOINTER_TO_INT (li->data); 
-			IGNORE_EINTR (close(fd));
+			VE_IGNORE_EINTR (close(fd));
 		}
 		g_slist_free (openfds);
 	} else {
@@ -1198,7 +1312,7 @@ gdm_close_all_descriptors (int from, int except, int except2)
 		}
 		for (i = from; i < max; i++) {
 			if G_LIKELY (i != except && i != except2)
-				IGNORE_EINTR (close(i));
+				VE_IGNORE_EINTR (close(i));
 		}
 	}
 }
@@ -1207,12 +1321,12 @@ int
 gdm_open_dev_null (mode_t mode)
 {
 	int ret;
-	ret = open ("/dev/null", mode);
+	VE_IGNORE_EINTR (ret = open ("/dev/null", mode));
 	if G_UNLIKELY (ret < 0) {
 		/* never output anything, we're likely in some
 		 * strange state right now */
 		gdm_signal_ignore (SIGPIPE);
-		IGNORE_EINTR (close (2));
+		VE_IGNORE_EINTR (close (2));
 		gdm_fail ("Cannot open /dev/null, system on crack!");
 	}
 
@@ -1331,10 +1445,19 @@ gdm_locale_from_utf8 (const char *text)
 }
 
 static GdmHostent *
+#ifdef ENABLE_IPV6
+fillout_addrinfo (struct addrinfo *res, struct sockaddr *ia, const char *name)
+#else
 fillout_hostent (struct hostent *he_, struct in_addr *ia, const char *name)
+#endif
 {
 	GdmHostent *he;
 
+#ifdef ENABLE_IPV6
+	gint i;
+	gint addr_count = 0;
+	struct addrinfo *tempaddrinfo;
+#endif
 	he = g_new0 (GdmHostent, 1);
 
 	he->addrs = NULL;
@@ -1343,6 +1466,56 @@ fillout_hostent (struct hostent *he_, struct in_addr *ia, const char *name)
 	/* Sometimes if we can't look things up, we could end
 	   up with a dot in the name field which would screw
 	   us up.  Weird but apparently possible */
+#ifdef ENABLE_IPV6
+	if (res != NULL && res->ai_canonname != NULL) {
+		he->hostname = g_strdup (res->ai_canonname);
+		he->not_found = FALSE;
+	} else {
+		he->not_found = TRUE;
+		if (name != NULL)
+			he->hostname = g_strdup (name);
+		else {
+			static char buffer6[INET6_ADDRSTRLEN], buffer[INET_ADDRSTRLEN];
+			const char *new = NULL;
+
+			if (ia->sa_family == AF_INET6) {
+				if (IN6_IS_ADDR_V4MAPPED (&((struct sockaddr_in6 *)ia)->sin6_addr))
+					new = inet_ntop (AF_INET, &(((struct sockaddr_in6 *)ia)->sin6_addr.s6_addr[12]), buffer, sizeof (buffer));
+
+				else
+					new = inet_ntop (AF_INET6, &((struct sockaddr_in6 *)ia)->sin6_addr, buffer6, sizeof (buffer6));
+			}
+			else if (ia->sa_family == AF_INET)
+				new = inet_ntop (AF_INET, &((struct sockaddr_in *)ia)->sin_addr, buffer, sizeof (buffer));
+
+			if (new)
+				he->hostname = g_strdup (new);
+			else
+				he->hostname = NULL;
+		}
+	}
+
+	tempaddrinfo = res;
+
+	while (res != NULL) {
+		addr_count ++;
+		res = res->ai_next;
+	}
+
+	he->addrs = g_new0 (struct sockaddr_storage, addr_count);
+	he->addr_count = addr_count;
+	res = tempaddrinfo;
+	for (i = 0; ;i++) {
+		if (res == NULL)
+			break;
+
+		if ((res->ai_family == AF_INET) || (res->ai_family == AF_INET6)) {
+			(he->addrs)[i] = *(struct sockaddr_storage *)(res->ai_addr);
+		}
+
+		res = res->ai_next;
+	}
+#else
 	if (he_ != NULL &&
 	    he_->h_name != NULL &&
 	    he_->h_name[0] != '\0' &&
@@ -1373,6 +1546,7 @@ fillout_hostent (struct hostent *he_, struct in_addr *ia, const char *name)
 			(he->addrs)[i] = *ia_;
 		}
 	}
+#endif
 	return he;
 }
 
@@ -1455,8 +1629,14 @@ jumpback_sighandler (int signal)
 GdmHostent *
 gdm_gethostbyname (const char *name)
 {
+#ifdef ENABLE_IPV6
+	struct addrinfo hints;
+	/* static because of Setjmp */
+	static struct addrinfo *result;
+#else
 	/* static because of Setjmp */
 	static struct hostent * he_;
+#endif
 	SETUP_INTERRUPTS_FOR_TERM_DECLS
 
 	/* the cached address */
@@ -1464,7 +1644,9 @@ gdm_gethostbyname (const char *name)
 	static time_t last_time = 0;
 	static char *cached_hostname = NULL;
 
+#ifndef ENABLE_IPV6
 	he_ = NULL;
+#endif
 
 	if (cached_hostname != NULL &&
 	    strcmp (cached_hostname, name) == 0) {
@@ -1478,12 +1660,30 @@ gdm_gethostbyname (const char *name)
 	if (Setjmp (signal_jumpback) == 0) {
 		do_jumpback = TRUE;
 		/* Find client hostname */
+#ifdef ENABLE_IPV6
+		memset (&hints, 0, sizeof (hints));
+		hints.ai_socktype = SOCK_DGRAM;
+		hints.ai_flags = AI_CANONNAME;
+
+		if (result) {
+			freeaddrinfo (result);
+			result = NULL;
+		}
+
+		getaddrinfo (name, NULL, &hints, &result);
+		do_jumpback = FALSE;
+	} else {
+               /* here we got interrupted */
+		result = NULL;
+	}
+#else
 		he_ = gethostbyname (name);
 		do_jumpback = FALSE;
 	} else {
 		/* here we got interrupted */
 		he_ = NULL;
 	}
+#endif
 
 	SETUP_INTERRUPTS_FOR_TERM_TEARDOWN
 
@@ -1491,17 +1691,34 @@ gdm_gethostbyname (const char *name)
 	cached_hostname = g_strdup (name);
 
 	gdm_hostent_free (he);
+#ifdef ENABLE_IPV6
+	he = fillout_addrinfo (result, NULL, name);
+#else
 	he = fillout_hostent (he_, NULL, name);
+#endif
 
 	last_time = time (NULL);
 	return gdm_hostent_copy (he);
 }
 
 GdmHostent *
-gdm_gethostbyaddr (struct in_addr *ia)
+#ifdef ENABLE_IPV6
+gdm_gethostbyaddr (struct  sockaddr_storage *ia)
+#else
+gdm_gethostbyaddr (struct sockaddr_in *ia)
+#endif
 {
+#ifdef ENABLE_IPV6
+	struct addrinfo hints;
+	/* static because of Setjmp */
+	static struct addrinfo *result = NULL;
+	struct sockaddr_in6 sin6;
+	struct sockaddr_in sin;
+	static struct in6_addr cached_addr6;
+#else
 	/* static because of Setjmp */
 	static struct hostent * he_;
+#endif
 	SETUP_INTERRUPTS_FOR_TERM_DECLS
 
 	/* the cached address */
@@ -1509,13 +1726,25 @@ gdm_gethostbyaddr (struct in_addr *ia)
 	static time_t last_time = 0;
 	static struct in_addr cached_addr;
 
+#ifndef ENABLE_IPV6
 	he_ = NULL;
+#endif
 
-	if (last_time != 0 &&
-	    memcmp (&cached_addr, ia, sizeof (struct in_addr)) == 0) {
-		/* don't check more then every 60 seconds */
-		if (last_time + 60 > time (NULL))
-			return gdm_hostent_copy (he);
+	if (last_time != 0) {
+#ifdef ENABLE_IPV6
+		if ((ia->ss_family == AF_INET6) && (memcmp (cached_addr6.s6_addr, ((struct sockaddr_in6 *) ia)->sin6_addr.s6_addr, sizeof (struct in6_addr)) == 0)) {
+			/* don't check more then every 60 seconds */
+			if (last_time + 60 > time (NULL))
+				return gdm_hostent_copy (he);
+		} else if (ia->ss_family == AF_INET)
+#endif
+		{
+			if (memcmp (&cached_addr, &(((struct sockaddr_in *)ia)->sin_addr), sizeof (struct in_addr)) == 0) {
+				/* don't check more then every 60 seconds */
+				if (last_time + 60 > time (NULL))
+					return gdm_hostent_copy (he);
+			}
+		}
 	}
 
 	SETUP_INTERRUPTS_FOR_TERM_SETUP
@@ -1523,20 +1752,73 @@ gdm_gethostbyaddr (struct in_addr *ia)
 	if (Setjmp (signal_jumpback) == 0) {
 		do_jumpback = TRUE;
 		/* Find client hostname */
-		he_ = gethostbyaddr ((gchar *) ia, sizeof (struct in_addr), AF_INET);
+
+#ifdef ENABLE_IPV6
+		memset (&hints, 0, sizeof (hints));
+		hints.ai_socktype = SOCK_DGRAM;
+		hints.ai_flags = AI_CANONNAME;
+
+		if (result) {
+			freeaddrinfo (result);
+			result = NULL;
+		}
+
+		if (ia->ss_family == AF_INET6) {
+			char buffer6[INET6_ADDRSTRLEN];
+
+			inet_ntop (AF_INET6, &((struct sockaddr_in6 *)ia)->sin6_addr, buffer6, sizeof (buffer6));
+
+			/* In the case of IPv6 mapped address strip the ::ffff: and lookup as an IPv4 address */
+			if (strncmp (buffer6, "::ffff:", 7) == 0) {
+				char *temp= (buffer6 + 7);
+				strcpy (buffer6, temp);
+			}
+			getaddrinfo (buffer6, NULL, &hints, &result);
+		}
+		else if (ia->ss_family == AF_INET) {
+			char buffer[INET_ADDRSTRLEN];
+
+			inet_ntop (AF_INET, &((struct sockaddr_in *)ia)->sin_addr, buffer, sizeof (buffer));
+
+			getaddrinfo (buffer, NULL, &hints, &result);
+		}
+
+		do_jumpback = FALSE;
+	} else {
+		/* here we got interrupted */
+		result = NULL;
+	}
+#else
+		he_ = NULL;
+		he_ = gethostbyaddr ((gchar *) &(ia->sin_addr), sizeof (struct in_addr), AF_INET);
 		do_jumpback = FALSE;
 	} else {
 		/* here we got interrupted */
 		he_ = NULL;
 	}
+#endif
 
 	SETUP_INTERRUPTS_FOR_TERM_TEARDOWN
 
-	cached_addr = *ia;
-
-	gdm_hostent_free (he);
-	he = fillout_hostent (he_, ia, NULL);
-
+#ifdef ENABLE_IPV6
+	if (ia->ss_family == AF_INET6) {
+		memcpy (cached_addr6.s6_addr, ((struct sockaddr_in6 *)ia)->sin6_addr.s6_addr, sizeof (struct in6_addr));
+		memset (&sin6, 0, sizeof (sin6));
+		memcpy (sin6.sin6_addr.s6_addr, cached_addr6.s6_addr, sizeof (struct in6_addr));
+		sin6.sin6_family = AF_INET6;
+		he = fillout_addrinfo (result, (struct sockaddr *)&sin6, NULL);
+	}
+	else if (ia->ss_family == AF_INET) {
+		memcpy (&(cached_addr.s_addr), &(((struct sockaddr_in *)ia)->sin_addr.s_addr), sizeof (struct in_addr));
+		memset (&sin, 0, sizeof (sin));
+		memcpy (&sin.sin_addr, &cached_addr, sizeof (struct in_addr));
+		sin.sin_family = AF_INET;
+		he = fillout_addrinfo (result, (struct sockaddr *)&sin, NULL);
+	}
+#else
+	cached_addr = ia->sin_addr;
+	he = fillout_hostent (he_, &(ia->sin_addr), NULL);
+#endif
 	last_time = time (NULL);
 	return gdm_hostent_copy (he);
 }
@@ -1557,8 +1839,13 @@ gdm_hostent_copy (GdmHostent *he)
 		cpy->addrs = NULL;
 	} else {
 		cpy->addr_count = he->addr_count;
+#ifdef ENABLE_IPV6
+		cpy->addrs = g_new0 (struct sockaddr_storage, he->addr_count);
+		memcpy (cpy->addrs, he->addrs, sizeof (struct sockaddr_storage) * he->addr_count);
+#else
 		cpy->addrs = g_new0 (struct in_addr, he->addr_count);
 		memcpy (cpy->addrs, he->addrs, sizeof (struct in_addr) * he->addr_count);
+#endif
 	}
 	return cpy;
 }
@@ -1583,38 +1870,11 @@ FILE *
 gdm_safe_fopen_w (const char *file)
 {
 	int fd;
-	IGNORE_EINTR (unlink (file));
-	fd = open (file, O_EXCL|O_CREAT|O_TRUNC|O_WRONLY
-#ifdef O_NOCTTY
-		   |O_NOCTTY
-#endif
-#ifdef O_NOFOLLOW
-		   |O_NOFOLLOW
-#endif
-		   , 0644);
-	if (fd < 0)
-		return NULL;
-	return fdopen (fd, "w");
-}
-
-/* like fopen with "a+" */
-FILE *
-gdm_safe_fopen_ap (const char *file)
-{
-	int fd;
-
-	if (access (file, F_OK) == 0) {
-		fd = open (file, O_APPEND|O_RDWR
-#ifdef O_NOCTTY
-			   |O_NOCTTY
-#endif
-#ifdef O_NOFOLLOW
-			   |O_NOFOLLOW
-#endif
-			  );
-	} else {
-		/* doesn't exist, open with O_EXCL */
-		fd = open (file, O_EXCL|O_CREAT|O_RDWR
+	FILE *ret;
+	VE_IGNORE_EINTR (unlink (file));
+	do {
+		errno = 0;
+		fd = open (file, O_EXCL|O_CREAT|O_TRUNC|O_WRONLY
 #ifdef O_NOCTTY
 			   |O_NOCTTY
 #endif
@@ -1622,10 +1882,50 @@ gdm_safe_fopen_ap (const char *file)
 			   |O_NOFOLLOW
 #endif
 			   , 0644);
+	} while G_UNLIKELY (errno == EINTR);
+	if (fd < 0)
+		return NULL;
+	VE_IGNORE_EINTR (ret = fdopen (fd, "w"));
+	return ret;
+}
+
+/* like fopen with "a+" */
+FILE *
+gdm_safe_fopen_ap (const char *file)
+{
+	int fd;
+	FILE *ret;
+
+	if (access (file, F_OK) == 0) {
+		do {
+			errno = 0;
+			fd = open (file, O_APPEND|O_RDWR
+#ifdef O_NOCTTY
+				   |O_NOCTTY
+#endif
+#ifdef O_NOFOLLOW
+				   |O_NOFOLLOW
+#endif
+				  );
+		} while G_UNLIKELY (errno == EINTR);
+	} else {
+		/* doesn't exist, open with O_EXCL */
+		do {
+			errno = 0;
+			fd = open (file, O_EXCL|O_CREAT|O_RDWR
+#ifdef O_NOCTTY
+				   |O_NOCTTY
+#endif
+#ifdef O_NOFOLLOW
+				   |O_NOFOLLOW
+#endif
+				   , 0644);
+		} while G_UNLIKELY (errno == EINTR);
 	}
 	if (fd < 0)
 		return NULL;
-	return fdopen (fd, "a+");
+	VE_IGNORE_EINTR (ret = fdopen (fd, "a+"));
+	return ret;
 }
 
 #ifdef RLIM_NLIMITS
@@ -1888,4 +2188,95 @@ gdm_strip_extension (const char *name, const char *extension)
 	}
 }
 
+void
+gdm_twiddle_pointer (GdmDisplay *disp)
+{
+	if (disp == NULL ||
+	    disp->dsp == NULL)
+		return;
+
+	XWarpPointer (disp->dsp,
+		      None /* src_w */,
+		      None /* dest_w */,
+		      0 /* src_x */,
+		      0 /* src_y */,
+		      0 /* src_width */,
+		      0 /* src_height */,
+		      1 /* dest_x */,
+		      1 /* dest_y */);
+	XSync (disp->dsp, False);
+	XWarpPointer (disp->dsp,
+		      None /* src_w */,
+		      None /* dest_w */,
+		      0 /* src_x */,
+		      0 /* src_y */,
+		      0 /* src_width */,
+		      0 /* src_height */,
+		      -1 /* dest_x */,
+		      -1 /* dest_y */);
+	XSync (disp->dsp, False);
+}
+
+static char *
+compress_string (const char *s)
+{
+	GString *gs = g_string_new (NULL);
+	const char *p;
+	gboolean in_whitespace = TRUE;
+
+	for (p = s; *p != '\0'; p++) {
+		if (*p == ' ' || *p == '\t') {
+			if ( ! in_whitespace)
+				g_string_append_c (gs, *p);
+			in_whitespace = TRUE;
+		} else {
+			g_string_append_c (gs, *p);
+			in_whitespace = FALSE;
+		}
+	}
+
+	return g_string_free (gs, FALSE);
+}
+
+
+char *
+gdm_get_last_info (const char *username)
+{
+	char *info = NULL;
+	const char *cmd = NULL;
+
+	if G_LIKELY (access ("/usr/bin/last", X_OK) == 0)
+		cmd = "/usr/bin/last";
+	else if (access ("/bin/last", X_OK) == 0)
+		cmd = "/bin/last";
+
+	if G_LIKELY (cmd != NULL) {
+		char *user_quoted = g_shell_quote (username);
+		char *newcmd;
+		FILE *fp;
+
+		newcmd = g_strdup_printf ("%s %s", cmd, user_quoted);
+
+		VE_IGNORE_EINTR (fp = popen (newcmd, "r"));
+
+		g_free (user_quoted);
+		g_free (newcmd);
+
+		if G_LIKELY (fp != NULL) {
+			char buf[256];
+			char *r;
+			VE_IGNORE_EINTR (r = fgets (buf, sizeof (buf), fp));
+			if G_LIKELY (r != NULL) {
+				char *s = compress_string (buf);
+				if ( ! ve_string_empty (s))
+					info = g_strdup_printf (_("Last login:\n%s"), s);
+				g_free (s);
+			}
+			VE_IGNORE_EINTR (pclose (fp));
+		}
+	}
+
+	return info;
+}
+
 /* EOF */
